Attribute VB_Name = "StepUpdates"
'  FILE:       StepUpdates.bas
'              Microsoft TPC-H Kit Ver. 2.7.0-1005
'              Copyright Microsoft, 2008
'              All Rights Reserved
'
'
'  PURPOSE:    This module contains the procedures that carry out the
'              inserts/updates/deletes of a step
'  Contact:    Reshma Tharamal (reshmat@microsoft.com)
'
Option Explicit

' Used to indicate the source module name when errors
' are raised by this module
Private Const mstrModuleName As String = "StepUpdates."

Public Enum CreateType
    Copy = 1
    Move
    Insert
End Enum
Private Sub DeleteStepRecord(strKey As String, _
        ArchivedFlag As Boolean, _
        ReorderFlag As Boolean, _
        Optional ByVal lngIdentifier As Long = 0)
    
    Dim lngStepId As Long
    Dim cStepToDelete As cStep
    
    On Error GoTo DeleteStepRecordErr
    
    If lngIdentifier = 0 Then
        lngStepId = MakeIdentifierValid(strKey)
    Else
        lngStepId = lngIdentifier
    End If
        
    Set cStepToDelete = gcSteps.QueryStep(lngStepId)
    
    ' Check if the step has to be archived
    If ArchivedFlag Then
        
        ' Update the archived flag on the step.
        cStepToDelete.ArchivedFlag = True
        gcSteps.Modify cStepToDelete
    Else
        
        ' Delete the step
        gcSteps.Delete cStepToDelete.StepId

        ' Call a procedure to delete all iterators
        ' for the step
        Call DeleteStepIterators(cStepToDelete)
    End If

    ' Call a procedure to delete/unload all constraint records
    ' for the step - if we are archiving the step, we need
    ' to unload the step, else just delete
    Call DeleteStepConstraints(cStepToDelete, _
            ArchivedFlag:=ArchivedFlag)
    
    If (ReorderFlag = True) Then
        ' Decrement the sequence number for all following steps by 1
        Call UpdateStepSequence(cStepToDelete, strKey, bInsert:=False)
    End If
    
    ' Delete the step from the tree and list views
    Call DeleteViewNode(frmMain.treStepView, frmMain.lvwStepView, _
            cStepToDelete.StepId, cStepToDelete.StepType, _
            cStepToDelete.WorkspaceId)

    Set cStepToDelete = Nothing
    
    Exit Sub
    
DeleteStepRecordErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    On Error GoTo 0
    Err.Raise vbObjectError + errStepDeleteFailed, _
            mstrModuleName & "DeleteStepRecord", _
            LoadResString(errStepDeleteFailed)
    
End Sub

Private Function FindSibling(ByVal cNewStep As cStep) As String

    Dim sParent As String
    Dim sNext As String
    Dim cTemp As cStep
    
    On Error GoTo FindSiblingErr
    gstrSource = mstrModuleName & "FindSibling"
    
    If cNewStep.ParentStepId = 0 Then
        sParent = LabelStep(cNewStep.WorkspaceId)
    Else
        sParent = MakeKeyValid(cNewStep.ParentStepId, gintManagerStep)
    End If
    
    If frmMain.treStepView.Nodes(sParent).Children > 0 Then
        sNext = frmMain.treStepView.Nodes(sParent).Child.Key
        
        Do
            Set cTemp = gcSteps.QueryStep(MakeIdentifierValid(sNext))
            If cTemp.SequenceNo = cNewStep.SequenceNo And cTemp.StepId <> cNewStep.StepId Then
                FindSibling = sNext
                Exit Function
            End If
            
            If sNext <> frmMain.treStepView.Nodes(sNext).LastSibling.Key Then
                sNext = frmMain.treStepView.Nodes(sNext).Next.Key
            Else
                sNext = gstrEmptyString
            End If
        Loop While Not StringEmpty(sNext)
        ' Could not find a sibling for this node
        FindSibling = gstrEmptyString
    Else
        FindSibling = gstrEmptyString
    End If
    
    Exit Function
    
FindSiblingErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    gstrSource = mstrModuleName & "FindSibling"
    On Error GoTo 0
    Err.Raise vbObjectError + errInsertStepFailed, _
            gstrSource, _
            LoadResString(errInsertStepFailed)
    
End Function
Public Sub InsertNewStepVersion(ByRef cStepToIncrement As cStep, _
    strOldVersion As String)
    
    ' Does all the processing necessary to create a step record
    ' with a new version on it. Each database update must be followed
    ' by a corresponding update to the step record in the array
    
    Dim strNewVersion As String
    Dim lngIndex As Long
    
    On Error GoTo InsertNewStepVersionErr
    
    strNewVersion = cStepToIncrement.VersionNo
    
    ' Update the version number on the step record and add it to the
    ' database. This will create a copy of the step record with a
    ' new version on it.
    lngIndex = gcSteps.QueryStepIndex(cStepToIncrement.StepId)
    cStepToIncrement.IndOperation = InsertOp
    
    Set gcSteps(lngIndex) = cStepToIncrement
    
    ' Call a procedure to create new versions of the constraint
    ' records for the step. It will also update the constraint
    ' array with the new step versions.
    gcConstraints.CreateNewConstraintVersion cStepToIncrement.StepId, _
            strNewVersion, strOldVersion, cStepToIncrement.StepType
    
    ' Call a procedure to create new versions of the iterator
    ' records for the step.
    cStepToIncrement.UpdateIteratorVersion
    
    ' Call a procedure to update all sub-steps of the passed in
    ' step, if any, with the new version number. It will also
    ' update the steps array.
    gcSteps.UpdateParentVersion cStepToIncrement.StepId, strNewVersion, strOldVersion, _
            cStepToIncrement.StepType
    
    Exit Sub
    
InsertNewStepVersionErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    On Error GoTo 0
    Err.Raise vbObjectError + errInsertNewStepVersionFailed, _
            gstrSource, _
            LoadResString(errInsertNewStepVersionFailed)
    
End Sub
Public Sub DeleteSubSteps(strRootNode As String, _
    ArchivedFlag As Boolean)
    ' Deletes/archives all the sub-step records for the passed in
    ' root node, depending on the blnArchive flag
    
    Dim nodCurrent As Node
    Dim nodNext As Node
    Dim strStepLabel As String
    
    On Error GoTo DeleteSubStepsErr
    
    strStepLabel = strRootNode
    
    If frmMain.treStepView.Nodes(strStepLabel).Children > 0 Then
        Set nodNext = frmMain.treStepView.Nodes(strStepLabel).Child
    End If
    
    ' While more step nodes exist
    While frmMain.treStepView.Nodes(strStepLabel).Children > 0
        
        Set nodCurrent = nodNext
        
        If nodCurrent.Children > 0 Then
            Set nodNext = nodCurrent.Child
        Else
            
            ' Set the temp pointer to the next sibling if it
            ' exists
            If nodCurrent.Index <> nodCurrent.LastSibling.Index Then
                Set nodNext = nodCurrent.Next
            ' Set the temp node pointer to the parent.
            ' Once we get to the steps label node, we are done
            ElseIf nodCurrent.Parent.Key <> strStepLabel Then
                Set nodNext = nodCurrent.Parent
            End If
        
            ' Since the current node does not have any children,
            ' it can be deleted
            ' Call a procedure to carry out the processing necessary
            ' to delete a step from the db and the loaded workspace
            ' Reorder sibling steps only for the root node
            Call DeleteStepRecord(strKey:=nodCurrent.Key, _
                    ArchivedFlag:=ArchivedFlag, ReorderFlag:=False)
        End If
    Wend

    Exit Sub
    
DeleteSubStepsErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    On Error GoTo 0
    Err.Raise vbObjectError + errDeleteSubStepsFailed, _
            mstrModuleName & "DeleteSubSteps", _
            LoadResString(errDeleteSubStepsFailed)
    
End Sub
Public Sub PopulatePositionDtls(cFrom As cStep, ByRef cTo As cStep)
    
    On Error GoTo PopulatePositionDtlsErr
    
    If Not cFrom Is Nothing Then
        cTo.WorkspaceId = cFrom.WorkspaceId
        cTo.ParentStepId = cFrom.ParentStepId
        cTo.ParentVersionNo = cFrom.ParentVersionNo
        cTo.StepLevel = cFrom.StepLevel
        cTo.SequenceNo = cFrom.SequenceNo
    End If
    
    Exit Sub
    
PopulatePositionDtlsErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    On Error GoTo 0
    Err.Raise vbObjectError + errNewStepFailed, _
            mstrModuleName & "PopulatePositionDtls", _
            LoadResString(errNewStepFailed)
    
End Sub

Private Function RunExistsForStep(ByVal lngStepIdentifier As Long, _
    ByVal strStepVersion As String) As Boolean
    ' Check if a run exists for the given version of the step
   
    Dim rstRun As Recordset
    Dim qyRun As DAO.QueryDef
    Dim strSql As String
    Dim cTempStr As cStringSM

    On Error GoTo RunExistsForStepErr
    gstrSource = mstrModuleName & "RunExistsForStep"
    
    Set cTempStr = New cStringSM

    ' Create a Recordset object to retrieve all steps for
    ' the given workspace
    strSql = "Select count(*) as run_count " & _
        " from run_step_details " & _
        " where step_id = [s_id] " & _
        " and version_no = [ver_no] "
    
    Set qyRun = dbsAttTool.CreateQueryDef(gstrEmptyString, strSql)
    qyRun.Parameters("s_id").Value = lngStepIdentifier
    qyRun.Parameters("ver_no").Value = strStepVersion
    
    Set rstRun = qyRun.OpenRecordset(dbOpenForwardOnly)
    
    Set cTempStr = Nothing

    If rstRun![run_count] > 0 Then
        RunExistsForStep = True
    Else
        RunExistsForStep = False
    End If
    rstRun.Close
    qyRun.Close

    Exit Function
    
RunExistsForStepErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    On Error GoTo 0
    Err.Raise vbObjectError + errRunExistsForStepFailed, _
            gstrSource, _
            LoadResString(errRunExistsForStepFailed)

End Function
Private Sub UpdateBranchLevel(cMoveStep As cStep, cOldParent As cStep, cNewParent As cStep)

    ' Handles all step level and step sequence updates when a branch is being moved
    
    Dim cBranch As cNavTree
    Dim strNextNode As String
    Dim lngIdentifier As Long
    Dim sRoot As String
    Dim iLevel As Integer
    Dim cCurStep As cStep
        
    On Error GoTo MoveBranchInTreeErr
    gstrSource = mstrModuleName & "MoveBranchInTree"
    
    sRoot = MakeKeyValid(cMoveStep.StepId, cMoveStep.StepType)
    
    ' Number of levels that the branch is moving up by
    If (cOldParent Is Nothing) Then
        If (cNewParent Is Nothing) Then
            iLevel = 0
        Else
            iLevel = cNewParent.StepLevel + 1
        End If
    ElseIf (cNewParent Is Nothing) Then
        iLevel = -(cOldParent.StepLevel + 1)
    Else
        iLevel = cNewParent.StepLevel - cOldParent.StepLevel
    End If
    
    If iLevel <> 0 Then
        Set cBranch = New cNavTree
        Set cBranch.Tree = frmMain.treStepView
        cBranch.Direction = gintBottomUp
        cBranch.RootNode = sRoot
        
        Do
            strNextNode = cBranch.NextNode
            If Not StringEmpty(strNextNode) Then
                lngIdentifier = MakeIdentifierValid(strNextNode)
                Set cCurStep = gcSteps.QueryStep(lngIdentifier)
                
                cCurStep.StepLevel = cCurStep.StepLevel + iLevel
                Call UpdateStepForMove(cCurStep)
            End If
        Loop While Not StringEmpty(strNextNode)
        
        Set cBranch = Nothing
    End If
    
    Exit Sub
    
MoveBranchInTreeErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    gstrSource = mstrModuleName & "MoveBranchInTree"
    On Error GoTo 0
    Err.Raise vbObjectError + errPasteNodeFailed, _
            gstrSource, _
            LoadResString(errPasteNodeFailed)

End Sub
Public Sub UpdateStep(ByRef cModifyStep As cStep)
    ' Carries out the operations necessary to update a
    ' step - all step properties have to be initialized
    ' before the step object is passed in
    
    Dim strStepNodeText As String
    
    On Error GoTo UpdateStepErr
    
    cModifyStep.Validate
    
    ' Check if the step updates are valid
    gcSteps.ValidateStep cModifyStep

    ' Insert a new parent step record with an incremented child
    ' version if runs exist for the parent step
    ' Ignoring the return value of the function
    Call CreateNewParent(cModifyStep, bDelete:=False)
        
    ' Call a procedure to insert a new version of the step if
    ' runs exist for the step
    If Not CreateNewStep(cModifyStep, IncrementX:=True) Then
        gcSteps.Modify cModifyStep
    End If
    
    ' Modify the tree and list view nodes, if any of the step
    ' properties have changed
    strStepNodeText = GetStepNodeText(cModifyStep)
    Call UpdateView(frmMain.treStepView, frmMain.lvwStepView, _
        cModifyStep.StepId, strStepNodeText, _
        cModifyStep, cModifyStep.StepType)

    ' Set the tag for the workspace node to indicate that
    ' it has been modified.
    Call WspModified(cModifyStep.WorkspaceId)
    
    Exit Sub

UpdateStepErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    gstrSource = mstrModuleName & "UpdateStep"
    On Error GoTo 0
    Err.Raise vbObjectError + errUpdateStepFailed, _
            gstrSource, _
            LoadResString(errUpdateStepFailed)
    
End Sub
Public Sub UpdateStepForMove(ByVal cModifyStep As cStep)
    ' Update step without checking for new versions
    
    On Error GoTo UpdateStepForMoveErr
    gstrSource = mstrModuleName & "UpdateStepForMove"
    
    gcSteps.Modify cModifyStep
    
    Exit Sub

UpdateStepForMoveErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    gstrSource = mstrModuleName & "UpdateStepForMove"
    On Error GoTo 0
    Err.Raise vbObjectError + errUpdateStepFailed, _
            gstrSource, _
            LoadResString(errUpdateStepFailed)
    
End Sub
Public Function CreateNewStep(ByRef cStepToCheck As cStep, _
        IncrementX As Boolean) As Boolean
    
    ' This function determines if the a new version of the step
    ' needs to be created. If so, it calls procedures to do the insert
    ' Returns True if a new version has been created
    
    Dim strOldVersion As String
    
    On Error GoTo CreateNewStepErr
    
    ' Check if any runs exist for the current version of the step
    If RunExistsForStep(cStepToCheck.StepId, cStepToCheck.VersionNo) Then

        ' We need to create a new step record with updated
        ' properties
        strOldVersion = cStepToCheck.VersionNo
        
        ' Call a function to determine the new version number for
        ' the step
        If IncrementX Then
            ' Increment the parent component of the version
            ' number by 1
            Call cStepToCheck.IncVersionX
        Else
            ' Increment the child component of the version
            ' number by 1
            Call cStepToCheck.IncVersionY
        End If
    
        ' Create a new step record and update the version number on
        ' all sub-steps for the current step
        Call InsertNewStepVersion(cStepToCheck, strOldVersion)
        
        CreateNewStep = True
    Else
        CreateNewStep = False
    End If
    
    Exit Function
    
CreateNewStepErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    On Error GoTo 0
    Err.Raise vbObjectError + errCreateNewStepFailed, _
            mstrModuleName & "CreateNewStep", _
            LoadResString(errCreateNewStepFailed)
    
End Function
Public Function CreateNewParent(ByRef cStepToCheck As cStep, bDelete As Boolean) As Boolean
    
    ' This function determines if the a new version of the step
    ' needs to be created. If so, it calls procedures to do the insert
    ' Returns True if a new version has been created
    
    Dim cParentStep As cStep
    
    On Error GoTo CreateNewParentErr
    
    ' Create a new version of the existing parent step, if a
    ' run exists for it
    If cStepToCheck.ParentStepId > 0 Then
        Set cParentStep = gcSteps.QueryStep(cStepToCheck.ParentStepId)
        
        ' Create a new version of the parent (if necessary)
        ' with the child component incremented by 1
        If CreateNewStep(cParentStep, IncrementX:=False) = True Then
            If Not bDelete Then
                ' Update the parent version number on the sub-step
                cStepToCheck.ParentVersionNo = cParentStep.VersionNo
            End If
            CreateNewParent = True
        End If
        Set cParentStep = Nothing
    Else
        CreateNewParent = False
    End If
    
    Exit Function
    
CreateNewParentErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    On Error GoTo 0
    Err.Raise vbObjectError + errCreateNewStepFailed, _
            mstrModuleName & "CreateNewParent", _
            LoadResString(errCreateNewStepFailed)
    
End Function
Public Sub InsertStep(ByVal cNewStep As cStep, _
    ByVal strAddStepTo As String, ByVal iCreate As CreateType, _
    Optional ByVal strMoveStepFrom As String = gstrEmptyString)
    ' Carries out the operations necessary to create a
    ' new step - all step properties have to be initialized
    ' before the step object is passed in
    
    Dim lngStepId As Long
    Dim strStepNodeText As String
    Dim sSibling As String
    
    On Error GoTo InsertStepErr
    
    If iCreate = Move Then
        BugAssert Not StringEmpty(strMoveStepFrom), "From key empty for move!"
        Call MoveBranch(cNewStep, strAddStepTo, strMoveStepFrom)
    Else
        ' Create a new version of the parent (if necessary)
        ' with the child component incremented by 1
        Call CreateNewParent(cNewStep, bDelete:=False)
        
        ' Create a dummy step object with the parent step id and
        ' version number equal to the one for the old parent
        If iCreate = Copy And WorkspaceSelected(strMoveStepFrom) <> cNewStep.WorkspaceId Then
            ' The step has been moved to a new workspace
            ' Delete all constraints for the step
            Call DeleteStepConstraints(cNewStep)
        End If
    
        ' Add the new step to the array
        gcSteps.Add cNewStep

        lngStepId = cNewStep.StepId
        strStepNodeText = GetStepNodeText(cNewStep)
        
        sSibling = FindSibling(cNewStep)
        If StringEmpty(sSibling) Then
            ' Load step values in the tree and list view controls
            Call AddNode(frmMain.treStepView, frmMain.lvwStepView, _
                    lngStepId, strStepNodeText, cNewStep.WorkspaceId, _
                    cNewStep.StepType, cNewStep, ParentKey:=strAddStepTo)
        Else
            ' Load step values in the tree and list view controls
            Call AddNode(frmMain.treStepView, frmMain.lvwStepView, _
                    lngStepId, strStepNodeText, cNewStep.WorkspaceId, _
                    cNewStep.StepType, cNewStep, SiblingKey:=sSibling)
        End If
    
        Call UpdateStepSequence(cNewStep, MakeKeyValid(cNewStep.StepId, cNewStep.StepType))
    
    End If
    
    ' Set the tag for the workspace node to indicate that
    ' it has been modified.
    Call WspModified(cNewStep.WorkspaceId)
    
    Exit Sub
    
InsertStepErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    gstrSource = mstrModuleName & "InsertStep"
    On Error GoTo 0
    Err.Raise vbObjectError + errInsertStepFailed, gstrSource, LoadResString(errInsertStepFailed)
    
End Sub
Public Sub MoveBranch(ByVal cNewStep As cStep, ByVal sTo As String, ByVal sFrom As String)
    ' This procedure contains all the processing necessary to
    ' move a branch from one node to another
    
    Dim cOldParent As cStep
    Dim cNewParent As cStep
    
    On Error GoTo MoveBranchErr
    gstrSource = mstrModuleName & "MoveBranch"
    
    If CircularReference(sTo, MakeKeyValid(cNewStep.StepId, cNewStep.StepType)) Then
        ShowError errCircularReference
        Exit Sub
    End If
    
    If Not StringEmpty(sFrom) And Not IsLabel(sFrom) Then
        Set cOldParent = gcSteps.QueryStep(MakeIdentifierValid(sFrom))
    End If
    
    If Not IsLabel(sTo) Then
        Set cNewParent = gcSteps.QueryStep(MakeIdentifierValid(sTo))
        
        cNewStep.ParentStepId = cNewParent.StepId
        cNewStep.ParentVersionNo = cNewParent.VersionNo
    Else
        cNewStep.ParentStepId = 0
        cNewStep.ParentVersionNo = gstrMinVersion
    End If
    
    If WorkspaceSelected(sFrom) <> cNewStep.WorkspaceId Then
        ' The step has been moved to a new workspace - delete all constraints for the step
        Call DeleteStepConstraints(cNewStep)
    End If
    
    ' Update the parent step id and version number on the
    ' step being moved - this must be done before new versions
    ' of the old and new parents are created, else the sub-steps
    ' will not be updated correctly.
    If Not CreateNewStep(cNewStep, IncrementX:=True) Then
        gcSteps.Modify cNewStep
    End If
        
    ' Check if a new version of the parent we are removing the
    ' branch from needs to be created
    If Not cOldParent Is Nothing Then
        Call CreateNewStep(cOldParent, IncrementX:=False)
    End If
    
    ' Check if a new version of the parent we are removing the
    ' branch to needs to be created
    If Not cNewParent Is Nothing Then
        Call CreateNewStep(cNewParent, IncrementX:=False)
    End If
    
    ' Check if the step_level of the branch being moved needs updating
    Call UpdateBranchLevel(cNewStep, cOldParent, cNewParent)
    
    ' Move the branch in the step tree
    Call MoveBranchInTree(MakeKeyValid(cNewStep.StepId, cNewStep.StepType), sTo)
            
    Exit Sub
    
MoveBranchErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    gstrSource = mstrModuleName & "MoveBranch"
    On Error GoTo 0
    Err.Raise vbObjectError + errMoveFailed, gstrSource, LoadResString(errMoveFailed)
    
End Sub
Private Sub MoveBranchInTree(strMoveStepKey As String, _
        strToNodeKey As String)
    ' Moves a branch from one location to another in the step tree
    ' Updating the parent property on the branch being moved is not enuf since we
    ' need to paste at a given location within the parent
    
    Dim cBranch As cNavTree
    Dim cNewNodeKeys As cVectorStr
    Dim cCurStep As cStep
    Dim nodTemp As Node
    Dim strNextNode As String
    Dim strNodeText As String
    Dim strParentKey As String
    Dim strAddKey As String
    Dim lngIdentifier As Long
    Dim lngIndex As Long
    Dim cRoot As cStep
    Dim sSibling As String
        
    On Error GoTo MoveBranchInTreeErr
    gstrSource = mstrModuleName & "MoveBranchInTree"
    
    Set cBranch = New cNavTree
    Set cBranch.Tree = frmMain.treStepView
    cBranch.Direction = gintTopDown
    cBranch.RootNode = strMoveStepKey
    
    Set cNewNodeKeys = New cVectorStr
    
    strNextNode = strMoveStepKey
    Do
        lngIdentifier = MakeIdentifierValid(strNextNode)
        Set cCurStep = gcSteps.QueryStep(lngIdentifier)
        
        ' Since the tree control won't accept duplicate
        ' values for the key, paste the nodes using a dummy
        ' suffix and then update the suffix later.
        strAddKey = MakeDummyKey(lngIdentifier)
        If strNextNode = strMoveStepKey Then
            Set cRoot = gcSteps.QueryStep(lngIdentifier)
            sSibling = FindSibling(cRoot)
        
            If StringEmpty(sSibling) Then
                strParentKey = strToNodeKey
            End If
        Else
            strParentKey = MakeDummyKey(MakeIdentifierValid(frmMain.treStepView.Nodes(strNextNode).Parent.Key))
        End If
        
        strNodeText = frmMain.treStepView.Nodes(strNextNode).Text
        ' Load step values in the tree and list view controls
        If StringEmpty(strParentKey) Then
            Call AddNode(frmMain.treStepView, frmMain.lvwStepView, _
                    lngIdentifier, strNodeText, cCurStep.WorkspaceId, _
                    cCurStep.StepType, cCurStep, SiblingKey:=sSibling, _
                    NodeKey:=strAddKey)
        Else
            Call AddNode(frmMain.treStepView, frmMain.lvwStepView, _
                    lngIdentifier, strNodeText, cCurStep.WorkspaceId, _
                    cCurStep.StepType, cCurStep, ParentKey:=strParentKey, _
                    NodeKey:=strAddKey)
        End If
                
        ' Add the newly created node key to a collection -
        ' these will be modified later
        cNewNodeKeys.Add strAddKey
        strNextNode = cBranch.NextNode
        
    Loop While Not StringEmpty(strNextNode)
    Set cBranch = Nothing
    
    ' Delete the branch from the tree
    Call UpdateStepSequence(cRoot, strMoveStepKey, bInsert:=False)
    frmMain.treStepView.Nodes.Remove strMoveStepKey
    
    ' And recreate the keys of the newly created nodes
    ' using valid suffixes
    For lngIndex = 0 To cNewNodeKeys.Count - 1
        strNextNode = cNewNodeKeys(lngIndex)
        Set cCurStep = gcSteps.QueryStep(MakeIdentifierValid(strNextNode))
        
        Set nodTemp = frmMain.treStepView.Nodes(strNextNode)
        nodTemp.Key = MakeKeyValid(cCurStep.StepId, cCurStep.StepType)
    Next lngIndex
    Set cNewNodeKeys = Nothing
    
    ' The same node is being both deleted and inserted!
    Call UpdateStepSequence(cRoot, strMoveStepKey, bInsert:=True)
    
    Exit Sub
    
MoveBranchInTreeErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    gstrSource = mstrModuleName & "MoveBranchInTree"
    On Error GoTo 0
    Err.Raise vbObjectError + errMoveFailed, gstrSource, _
            LoadResString(errMoveFailed)
    
End Sub
Private Function MakeDummyKey(lngIdentifier As Long) As String

    ' We need to re-create the branch in a new location.
    ' But the tree control won't accept duplicate values
    ' for the key. So we paste the nodes using a dummy
    ' suffix and then update the suffix later.
    MakeDummyKey = gstrDummyPrefix & Trim$(Str(lngIdentifier))
    
End Function
Public Sub CreateStep(ByVal cNewStep As cStep, _
        ByVal lngWorkspaceId As Long, ByVal strFrom As String, ByVal bCopy As Boolean)
    ' Does all the processing necessary to create a new record
    ' with all properties equal to the passed in step except
    ' for the position details - they will be initialized
    ' by this procedure depending on the node selected by the user

    Dim strParentNodeKey As String
    Dim cStepToCreate As cStep
    Dim vntSubSteps As Variant
    
    On Error GoTo CreateStepErr
    gstrSource = mstrModuleName & "CreateStep"
    
    ' Initialize the new step record to the one passed in
    Set cStepToCreate = cNewStep.Clone
    
    vntSubSteps = gcSteps.SubSteps(cStepToCreate.StepId, cStepToCreate.VersionNo)
    If Not IsEmpty(vntSubSteps) And cStepToCreate.WorkspaceId <> lngWorkspaceId Then
        ' We cannot handle branch moves to a new workspace
        ' currently - show an error
        On Error GoTo 0
        ShowError errMoveBranchInvalid
        Err.Raise vbObjectError + errMoveBranchInvalid, gstrSource, _
                LoadResString(errMoveBranchInvalid)
    End If
    
    ' Initialize the workspace identifier and a new step identifier
    cStepToCreate.WorkspaceId = lngWorkspaceId

    If cStepToCreate.StepType = gintGlobalStep Then
        ' Note: The workspace id in the step object has already been
        ' initialized above.
        strParentNodeKey = FindStepPositionDetails(cStepToCreate, _
                cStepToCreate.StepType)
    Else
        ' Now determine the properties that will determine the position
        ' of the step in the hierarchy
        strParentNodeKey = ItemSelectedInTree(frmMain.treStepView)
        If GetStepPosition(cStepToCreate, strParentNodeKey) = False Then
            Set cStepToCreate = Nothing
            Exit Sub
        End If
    End If
    
    If bCopy And IsEmpty(vntSubSteps) Then
        Call InsertStep(cStepToCreate, strParentNodeKey, Copy, strFrom)
    Else
        Call InsertStep(cStepToCreate, strParentNodeKey, Move, strFrom)
    End If
    
    If bCopy Then
        ' Call a procedure to copy all pre and post-execution
        ' constraint records
        Call CopyConstraintRecords(cNewStep, cStepToCreate)
'
'        ' Call a procedure to copy all iterators for the step
'        Call CopyIterators(cNewStep, cStepToCreate)
    End If
    
    Exit Sub
    
CreateStepErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    On Error GoTo 0
    gstrSource = mstrModuleName & "CreateStep"
    Err.Raise vbObjectError + errCreateStepFailed, gstrSource, LoadResString(errCreateStepFailed)
    
End Sub
Public Sub StepDelete(Optional ByVal DeleteStep As cStep, _
        Optional StepKey As String = gstrEmptyString, _
        Optional ByVal Identifier As Long = 0)
    ' This procedure will handle deletes of global, manager and
    ' worker steps
    
    Dim lngStepId As Long
    Dim cStepToDelete As cStep
    Dim cNewParentStep As cStep
    Dim strOldVersion As String
    Dim blnArchived As Boolean
    
    On Error GoTo StepDeleteErr
    gstrSource = mstrModuleName & "StepDelete"
    
    If Not DeleteStep Is Nothing Then
        Set cStepToDelete = DeleteStep
        lngStepId = cStepToDelete.StepId
        StepKey = MakeKeyValid(lngStepId, cStepToDelete.StepType)
    ElseIf Identifier > 0 Then
        lngStepId = Identifier
        Set cStepToDelete = gcSteps.QueryStep(lngStepId)
        StepKey = MakeKeyValid(lngStepId, cStepToDelete.StepType)
    ElseIf StepKey <> gstrEmptyString Then
        lngStepId = MakeIdentifierValid(StepKey)
        Set cStepToDelete = gcSteps.QueryStep(lngStepId)
    Else
        On Error GoTo 0
        Err.Raise vbObjectError + errMandatoryParameterMissing, _
                gstrSource, _
                LoadResString(errMandatoryParameterMissing)
    End If
        
    ' Check if the step has any sub-steps and if all of
    ' them should be moved
    If Not ConfirmDeleteOrMove(DeleteStep, StepKey) Then
        Exit Sub
    End If
    
    ' Check if any runs exist for the current version of the step
    If RunExistsForStep(cStepToDelete.StepId, cStepToDelete.VersionNo) Then
        ' If a run exists for the step being deleted, then the
        ' step record is not deleted from the database, but the
        ' archived flag is set to True. Moreover if the step is a
        ' sub-step, then we create a new version of the parent step.
        blnArchived = True
    ElseIf cStepToDelete.IsNewVersion Then
        cStepToDelete.VersionNo = cStepToDelete.OldVersionNo
        cStepToDelete.IndOperation = UpdateOp
        blnArchived = True
    Else
        ' No runs existed for the step. So, first delete all sub-steps
        ' associated with the step, if any and then delete the step
        blnArchived = False
    End If
        
    ' First update the archived flag on all sub-steps
    ' for the step and on the step. This has to be done
    ' before creating a new version of the parent (if necessary )
    ' since the version numbers of all sub-steps will be updated
    ' when a new version of the parent is created
    Call DeleteSubSteps(StepKey, ArchivedFlag:=blnArchived)
    
    Call DeleteStepRecord(StepKey, ArchivedFlag:=blnArchived, _
            ReorderFlag:=True, lngIdentifier:=lngStepId)
    
    If blnArchived Then
        Call CreateNewParent(cStepToDelete, bDelete:=True)
    End If
    
    ' Set the tag for the workspace node to indicate that
    ' it has been modified.
    Call WspModified(cStepToDelete.WorkspaceId)
    
    Set cStepToDelete = Nothing
    
    Exit Sub
    
StepDeleteErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    gstrSource = mstrModuleName & "StepDelete"
    On Error GoTo 0
    Err.Raise vbObjectError + errStepDeleteFailed, _
            gstrSource, _
            LoadResString(errStepDeleteFailed)
    
End Sub
Public Sub UpdateStepSequence(cStepRecord As cStep, sKey As String, _
        Optional ByVal bInsert As Boolean = True)
    ' Called to update the positions of all following steps when a step is
    ' inserted/deleted.
    Dim cUpdateStep As cStep
    Dim lngStepId As Long
    Dim sSibKey As String
    
    On Error GoTo UpdateStepSequenceErr
    
    sSibKey = sKey
    
    Do
        If sSibKey <> frmMain.treStepView.Nodes(sKey).LastSibling.Key Then
    
            sSibKey = frmMain.treStepView.Nodes(sSibKey).Next.Key
            lngStepId = MakeIdentifierValid(sSibKey)
            Set cUpdateStep = gcSteps.QueryStep(lngStepId)
            
            ' Modify the sequence # on the step record and call a procedure to carry
            ' out the update
            cUpdateStep.SequenceNo = IIf(bInsert, _
                    cUpdateStep.SequenceNo + 1, cUpdateStep.SequenceNo - 1)
            Call UpdateStepForMove(cUpdateStep)
        Else
            sSibKey = gstrEmptyString
        End If
    Loop While Not StringEmpty(sSibKey)

    Exit Sub
    
UpdateStepSequenceErr:
    ' Log the error code raised by Visual Basic
    Call LogErrors(Errors)
    gstrSource = mstrModuleName & "UpdateStepSequence"
    On Error GoTo 0
    Err.Raise vbObjectError + errUpdateStepFailed, _
            gstrSource, _
            LoadResString(errUpdateStepFailed)

End Sub


